/*
    ftremote - remote serial/bitbang for FT232 and similar chips
    Copyright (C) 2002-2009  Ladislav Vaiz <ok1zia@nagano.cz>
    and Even Nermerson (ftdi_new and free_ftdi) and other
    authors of libftdi

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#ifdef HAVE_LIBFTDI

struct ftdi_context *ftdi = NULL;
int ftdi_vid, ftdi_pid;
char *ftdi_serial;
int ftdi_chunksize;
int ftdi_timer_id = 0;

int init_ftdi(){
    ftdi = NULL;
    ftdi_open(1);
    return 0;
}


int ftdi_open(int verbose){
    int ret, i, ftdi_baudrate, ftdi_bitmode, ftdi_mask, ftdi_dummy;
    double samplerate;
            
    ftdi_vid = rc_int("ftdi_vid", 0xa600); /* USB vendor ID */
    ftdi_pid = rc_int("ftdi_pid", 0xe111); /* USB product ID */
    ftdi_serial = rc_str("ftdi_serial", ""); /* FTDI serial number */
    ftdi_baudrate = rc_int("ftdi_baudrate", 300); /* FTDI baud rate (300-3000000) */
    ftdi_bitmode = rc_int("ftdi_bitmode", 1); /* FTDI bitmode (0=serial, 1=bitbang etc) */
    ftdi_mask = rc_int("ftdi_mask", 0x00); /* FTDI bit mask (0=input, 1=output) */
    ftdi_chunksize = rc_int("ftdi_chunksize", 5); /* FTDI chunk size in bytes (5) */
    if (ftdi_chunksize > 1024) ftdi_chunksize = 1024;
    ftdi_dummy = rc_int("ftdi_dummy", 0); /* Dummy mode without FTDI device */
    
    
    if (ftdi_dummy){
        if (opt_s){
            dbg("installing dummy timer\n");
            ftdi_timer_id = install_timer(1000, ftdi_dummy_timer_read, CBA0);
        }
        return 0;
    }

    if (ftdi != NULL) return 0;

    if (!ftdi_vid) {
        error("Refuse to open VID=%04x, probably incorrect configuration.\n", ftdi_vid);
        return 1;
    }
    dbg("ftdi_open %04x:%04x  server=%d\n", ftdi_vid, ftdi_pid, opt_s);
    

    ftdi = ftdi_new();
    if (!ftdi){
        if (verbose) error("Can't init ftdi library for ftdi");
        return 1;
    }
    if (ftdi_serial && *ftdi_serial)
        ret=ftdi_usb_open_desc(ftdi, ftdi_vid, ftdi_pid, NULL, ftdi_serial);
    else
        ret=ftdi_usb_open(ftdi, ftdi_vid, ftdi_pid);
    
    if (ret){
        if (verbose) {
            error("Can't open ftdi device %04x:%04x, error=%d %s\n", ftdi_vid, ftdi_pid, ret, ftdi->error_str);
            if (ret==-8) error("Maybe run program as root\n");
        }
        free_ftdi();
        return ret;
    }
    
    ret=ftdi_set_baudrate(ftdi, ftdi_baudrate);
    if (ret){
        if (verbose) error("Can't set baudrate for ftdi, error=%d %s\n", ret, ftdi->error_str);
        free_ftdi();
        return ret;
    }
    ret=ftdi_set_bitmode(ftdi, ftdi_mask, ftdi_bitmode);
    dbg("ftdi_set_bitmode(0x%02x, %d)=%d\n", ftdi_mask, ftdi_bitmode, ret);
    //ret=ftdi_set_bitmode(ftdi, 0xff, BITMODE_BITBANG);
    if (ret){
        if (verbose) error("Can't enable bitbang, error=%d %s\n", ret, ftdi->error_str);
        free_ftdi();
        return ret;
    }
    
    ret=ftdi_write_data_set_chunksize(ftdi, 50);
    if (ret){
        if (verbose) error("Can't set chunk size, error=%d %s\n", ret, ftdi->error_str);
        free_ftdi();
        return ret;
    }
    
    if (!opt_s){
        ftdi_timer_id = install_timer(1000, ftdi_timer_read, CBA0);
    }

    if (opt_w && opt_r){
        int i, l=0, sec, usec;
        char buf[1024];
        double cas;
        struct timeval start, stop;
        
        ST_START;
        buf[0]=0;
        for(i=0;i<100;i++){
            buf[0]=i&1?0xff:0;
            ret=ftdi_write_data(ftdi, buf, 1);
            ret=ftdi_read_data(ftdi, buf,1);
        }
        ST_STOP;
        exit(0);
        
    }

    if (opt_w){
        int i, l=0, sec, usec;
        char buf[1024];
        double cas;
        struct timeval start, stop;
        
        for (i=0; i<1024;i++) buf[i]=i&1?0xff:0;
        
        gettimeofday(&start, NULL);
        while(1){
            ret=ftdi_write_data(ftdi, buf, 1024);
            if (ret!=1024) {
                error("Can't write, error=%d %s\n", ret, ftdi->error_str);
                exit(0);
            }
            l+=ret;
            gettimeofday(&stop, NULL);
            if (stop.tv_sec < start.tv_sec + 10) continue;

            usec=stop.tv_usec-start.tv_usec;
            sec=stop.tv_sec-start.tv_sec;
            if (usec<0){usec+=1000000;sec--;}
            cas = sec + usec/1000000.0;
            printf("%d bytu za %3.1f sec, %3.0f bytes/sec\n", l, cas, l/cas);
            l=0;
            gettimeofday(&start, NULL);
        }
        
    }

    if (opt_r){
        int i, l=0, sec, usec;
        char buf[1024];
        double cas;
        struct timeval start, stop;
        
        gettimeofday(&start, NULL);
        while(1){
            ret=ftdi_read_data(ftdi, buf, 1024);
            if (ret!=1024) {
                error("Can't read, error=%d %s\n", ret, ftdi->error_str);
                exit(0);
            }
            l+=ret;
            gettimeofday(&stop, NULL);
            if (stop.tv_sec < start.tv_sec + 10) continue;

            usec=stop.tv_usec-start.tv_usec;
            sec=stop.tv_sec-start.tv_sec;
            if (usec<0){usec+=1000000;sec--;}
            cas = sec + usec/1000000.0;
            printf("%d bytu za %3.1f sec, %3.0f bytes/sec\n", l, cas, l/cas);
            l=0;
            gettimeofday(&start, NULL);
        }
        
    }

/*    while(1){
        char s[2];
        s[0]=0xff;
        ret=ftdi_write_data(ftdi, s, 1);
        dbg("ret=%d\n", ret);
        sleep(2);
        s[1]=0x00;
        ret=ftdi_write_data(ftdi, s, 1);
        dbg("ret=%d\n", ret);
        sleep(2);
    } */
//    thread = SDL_CreateThread(ftdi_thread_func, ds);
    /*dbg("FTDI type %d\n", ftdi->type);    */
    
    return 0;
}

int free_ftdi(){
//    SDL_KillThread(thread);
    if (ftdi!=NULL){
        ftdi_free(ftdi);
        ftdi=NULL;
    }
    if (ftdi_timer_id) {
        kill_timer(ftdi_timer_id);
        ftdi_timer_id = 0;
    }
    return 0;
}


void ftdi_dummy_timer_read(union cba_t cba){
    static unsigned char counter = 0;

    if (!conn) {
        dbg("ftdi_dummy_timer_read conn==NULL\n");
    }else{
        write(conn->sock, &counter, 1);
        if (opt_d >= 3) dbg("ftdi_dummy_timer_read 0x%02x\n", counter);
        counter++;
    }
    ftdi_timer_id = install_timer(100, ftdi_dummy_timer_read, CBA0);
}

void ftdi_timer_read(union cba_t cba){
    char buf[1024], delta[1024], oldc;
    int ret, wr, i, di;

    if (!conn) {
        dbg("ftdi_timer_read conn==NULL\n");
    }else{
        ftdi_read_data(ftdi, &oldc, 1); // ignore return code
        while (1){
            ret = ftdi_read_data(ftdi, buf, ftdi_chunksize);
            if (ret < 0){
                error("Can't read from ftdi, error=%d %s", ret, ftdi->error_str);
                return;
            }
            for (i=0; i<ret; i++) buf[i]=buf[i]==-1?0xff:0x00;

            di=0;
            for (i=0; i<ret; i++){
                if (buf[i]==oldc) continue;
                delta[di++]=oldc=buf[i];
            }

            if (opt_d >= 4){
                dbg("ftdi_timer_read ");
                for (i=0; i<ret; i++) dbg("%02x ", (unsigned char)buf[i]);
                dbg("\n");
            }else if (opt_d >= 3){
                dbg("ftdi_timer_read delta ");
                for (i=0; i<di; i++) dbg("%02x ", (unsigned char)delta[i]);
                dbg("\n");
            }else{
                dbg("ftdi_timer_read len=%d\n", ret);
            }

            if (ret == 0) break; // nic se neprecetlo
            if (di == 0) continue; // nemame co posilat
            wr = write(conn->sock, delta, di);
            if (wr != di) {
                error("Can't write to socket %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
                tcp_disconnect(conn, "Can't write to socket");
                return;
            }
        }
    }
    ftdi_timer_id = install_timer(1000, ftdi_timer_read, CBA0);
}


#ifndef HAVE_FTDI_NEW

/* compatibility with libftdi < 0.11 */
/* code from libftdi 0.11 */

struct ftdi_context *ftdi_new()
{
    struct ftdi_context * ftdi = (struct ftdi_context *)malloc(sizeof(struct ftdi_context));

    if (ftdi == NULL) {
        return NULL;
    }

    if (ftdi_init(ftdi) != 0) {
        free(ftdi);
        return NULL;
    }

    return ftdi;
}



void free_ftdi(struct ftdi_context *ftdi)
{
    ftdi_deinit(ftdi);
    free(ftdi);
}

#endif

#define DUMP_EEPROMx

void usb_info(void){
    int ret;
    struct ftdi_context ftdi;
    struct ftdi_device_list *list, *dev;
    int ids[] = {0x0403, 0x6001, 
                 0xa600, 0xE110,
                 0xa600, 0xE111,
                 0xa600, 0xE112,
                 0};
    int idi;
    char manufacturer[128], description[128], serial[128];
#ifdef DUMP_EEPROM
    unsigned char eeprom_buf[128];
#endif
    
    printf("\n  usb_info:\n");
    ret = ftdi_init(&ftdi); 
    if (ret < 0) {
        printf("error calling ftdi_init, error=%d %s", ret, ftdi.error_str);
        return;
    }
    for (idi = 0; ids[idi] != 0; idi += 2){
        ret = ftdi_usb_find_all(&ftdi, &list, ids[idi], ids[idi+1]);
        if (ret < 0) {
            printf("error calling ftdi_usb_find_all, error=%d %s", ret, ftdi.error_str);
            return;
        }
        for (dev = list; dev != NULL; dev = dev->next){
            printf("%04x:%04x  ", dev->dev->descriptor.idVendor, dev->dev->descriptor.idProduct);
            ret = ftdi_usb_get_strings(&ftdi, dev->dev, manufacturer, 128, description, 128, serial, 128);
            if (ret < 0){
                printf("Can't get strings, error=%d %s", ret, ftdi.error_str);
            }else{
                printf("%s %s ", manufacturer, description);
                if (strlen(serial)>0 && serial[0]!='?')
                    printf("%s  ", serial);
            }
            printf("\n");
#ifdef DUMP_EEPROM
            ret = ftdi_usb_open_dev(&ftdi, dev->dev);
            if (ret < 0){
                printf("Can't open device, error=%d %s", ret, ftdi.error_str);
            }else{
                ret = ftdi_read_eeprom(&ftdi, eeprom_buf);
                if (ret < 0){
                    printf("Can't open device, error=%d %s", ret, ftdi.error_str);
                }else{
                    int i;
                    unsigned short chksum_in_eeprom, chksum_computed;
                    for (i=0; i<128; i++){
                        printf("%02x ", (unsigned char)eeprom_buf[i]);
                        if (i%16==15) printf("\n");
                        
                    }
                    chksum_in_eeprom = eeprom_buf[0x7e] + 256 * eeprom_buf[0x7f];
                    chksum_computed = ftdi_checksum(eeprom_buf, 0x80);
                    printf("Checksum eeprom=%04x computed=%04x ", chksum_in_eeprom, chksum_computed);
                }
                ftdi_usb_close(&ftdi);
            }
            printf("\n");
#endif
        }
    }
    ftdi_list_free(&list);
    ftdi_deinit(&ftdi);
}

#endif
