/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#define PROT_VER "1"

int tcpsock;
struct conn *conn;

int init_tcp(){
	int on, tcpport;
	char errbuf[1024];
	struct sockaddr_in sin;
    
    tcpsock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    /*dbg("   tcpsock=%d\n", net->tcpsock);*/
	if (tcpsock < 0){
        error("Can't create socket for tcp - %s\n", STRERR(errno));
		return -1;
	}

#ifndef __CYGWIN__    
    on=1;
    if (setsockopt(tcpsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))){
        error("Can't set SO_REUSEADDR - %s\n", STRERR(errno));
        goto x;
    }
#endif    
    
    if (fcntl(tcpsock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK - %s\n", STRERR(errno));
        goto x;
    }
        
    tcpport=rc_int("listen", 7300); /* TCP port to listen */
	memset(&sin, 0, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(tcpport);
	sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(tcpsock, (struct sockaddr *)&sin, sizeof(sin))){
		error("Can't bind TCP port %d - %s\n", tcpport, STRERR(errno)); 
		goto x;
	}

    if (listen(tcpsock, 10)){
        error("Can't listen on socket %d, tcp port %d - %s\n", tcpsock, tcpport, STRERR(errno));
        goto x;
    }

    set_handlers(tcpsock, tcp_accept_handler, NULL, tcp_exception_handler, CBA0);

    
    return 0;
x:;
    close(tcpsock);
    return -1;
}

void tcp_accept_handler(union cba_t cba){
    int sock;
    struct sockaddr_in sin;
    socklen_t socklen;
    char *c;
    char salt[100];
    
    
    
    /*dbg("tcp_accept_handler()\n");*/

    socklen = sizeof(sin);
    sock = accept(tcpsock, (struct sockaddr *)&sin, &socklen);
    if (!socklen || sock<0) return;

    dbg("s%d: accepted %s:%d\n", sock, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));

    conn = mem_alloc(sizeof(struct conn));
	bzero(conn, sizeof(struct conn));
    
    conn->sin.sin_family      = AF_INET;
    conn->sin.sin_addr.s_addr = sin.sin_addr.s_addr;
    conn->sin.sin_port        = sin.sin_port;

    
    conn->sock = sock;
    //tcp_set_state(conn, TCP_CONNECTED);
    
    set_handlers(conn->sock, tcp_read_handler, 
            NULL, tcp_exception_handler, (union cba_t)conn);

}


void tcp_connect(char *hostname, int port){
    int ret;
	char errbuf[1024];
    struct hostent *he;
        

    he = gethostbyname(hostname);
    if (!he){
        error("Can't resolve %s\n", hostname);
        return;
    }
    
    conn = mem_alloc(sizeof(struct conn));
	bzero(conn, sizeof(struct conn));

    conn->sin.sin_family      = AF_INET;
    memcpy(&conn->sin.sin_addr, he->h_addr_list[0], he->h_length);
    conn->sin.sin_port        = htons(port);
    
    
    
/*    dbg("tcp->connect %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));  */
   
    conn->sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
/*    dbg("socket returns %d \n", conn->sock);*/
    
    if (fcntl(conn->sock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK\n");
        return;
    }
    
    ret = connect(conn->sock, (struct sockaddr *) &conn->sin, sizeof(struct sockaddr_in));
/*    dbg("connect returns %d (%d) '%s'\n", ret, errno, strerror_r(errno));*/
    if (ret==0){
        set_handlers(conn->sock, tcp_read_handler, NULL, tcp_exception_handler, (union cba_t)conn);
        //tcp_set_state(conn, TCP_CONNECTED);
        after_connect(conn);
        return;
    }
    if (ret<0){
        if (errno==EALREADY || errno==EINPROGRESS){
            //tcp_set_state(conn, TCP_CONNECTING);
            set_handlers(conn->sock, NULL, tcp_connected_handler, tcp_exception_handler, (union cba_t)conn);
            dbg("s%d: connecting to %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
            return;
        }  
        error("connect error %s\n", STRERR(errno));
        tcp_disconnect(conn, "connect error");
    }
}

void tcp_connected_handler(union cba_t cba){
    struct conn *conn;
    int ret, err, cnt;
    socklen_t optlen;
    char errbuf[1024], s[1300];
    
    
    conn = cba.conn;
/*    dbg("tcp_connected_handler %s:%d\n", inet_ntoa(conn->sin.sin_addr),
 *    ntohs(conn->sin.sin_port));*/
    
    if (fcntl(conn->sock, F_SETFL, 0)){
        tcp_disconnect(conn, "Can't clear O_NONBLOCK");
        return;
    }

    cnt=10;
gsopt:;    
    optlen = sizeof(err);
    ret = getsockopt(conn->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &optlen);
    dbg("   getsockopt(SO_ERROR) returns %d err is %d\n", ret, err);
    if (ret!=0){
        tcp_disconnect(conn, "getsockopt returns !=0");
        return;
    }
    if (err!=0){
        sprintf(s, "getsockopt returns error %d - %s", err, STRERR(errno));
        if (errno==EINPROGRESS && --cnt>0) goto gsopt;
        error("not connected %s:%d - %s\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port), STRERR(errno));
        tcp_disconnect(conn, s);
        return;
    }
    
    error("connected %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    
    if (fcntl(conn->sock, F_SETFL, O_NONBLOCK)){
        tcp_disconnect(conn, "Can't set O_NONBLOCK");
        return;
    }
    //tcp_set_state(conn, TCP_CONNECTED);
    set_handlers(conn->sock, tcp_read_handler, 
            NULL, tcp_exception_handler, (union cba_t)conn);
    after_connect(conn);
}

void after_connect(struct conn *conn){
    char *c;
    dbg("s%d: connected to %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
}

void tcp_disconnect(struct conn *conn, char *reason){
    struct conn *cc;

    if (conn->disconnecting) return;
    conn->disconnecting=1;

	dbg("s%d: disconnected %s:%d because of %s\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port), reason);
    set_handlers(conn->sock, NULL, NULL, NULL, CBA0);
    if (conn->timer) kill_timer(conn->timer);

    close(conn->sock);
	mem_free(conn);
}

void tcp_read_handler(union cba_t cba){
    struct conn *conn;            
    char buf[1030], *d;
    int ret, err, len, i, wr;
    char errbuf[1030];

    conn = cba.conn;
    ret=read(conn->sock, buf, 5);
    err=errno;
	if (ret<0){
		error("Error reading socket - %s\n", STRERR(errno));
		tcp_disconnect(conn, "read returns <0");
		return;
	}
	if (ret==0){
		error("Got disconnect\n");
		tcp_disconnect(conn, "got disconnect");
        free_ftdi();
		return;
	}

    if (opt_d>=3) {
        dbg("tcp_read_handler ");
        for (i=0;i<ret;i++) dbg("%02x ", (unsigned char)buf[i]);
        dbg("\n");
    }else{
        dbg("tcp_read_handler len=%d\n", ret);
    }

    if (!ftdi){
        dbg("discarding %d bytes\n", ret);
    }else{
        wr = ftdi_write_data(ftdi, buf, ret);
        dbg("wr=%d\n", wr);
        if (wr < 0){
            error("Can't write to ftdi, error=%d %s\n", wr, ftdi->error_str);
            tcp_disconnect(conn, "Can't write to ftdi");
        }
    }

	return;
}

void tcp_exception_handler(union cba_t cba){
    struct conn *conn;
    
    conn = cba.conn;
    dbg("s%d: tcp_exception_handler %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port)); 
    error("disconnected %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port)); 
    
    tcp_disconnect(conn, "socket exception");
}

int tcp_send_state(struct conn *conn){
	
	return 0;
}



void kill_conn_timer(void *xxx){
    struct conn *conn;

    conn = (struct conn *)xxx;
    dbg("s%d: kill_conn_timer %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    conn->timer = 0;
    tcp_disconnect(conn, "connection or ack timeout");
}

void remove_conn_timer(void *xxx){
    struct conn *conn;

    conn = (struct conn *)xxx;
/*    dbg("remove_conn_timer %s:%d\n", inet_ntoa(conn->sin.sin_addr),
 *    ntohs(conn->sin.sin_port));*/
    conn->timer = 0;

    free_conn(conn);
    mem_free(conn);
}

void free_conn(struct conn *conn){
    dbg("s%d: free_conn\n", conn->sock);
    set_handlers(conn->sock, NULL, NULL, NULL, CBA0);
    if (conn->sock>=0) close(conn->sock);
    conn->sock=-1;
//    tcp_set_state(conn, TCP_DEAD);
}


